"""
Builder

Classes for programmatically constructing and writing out a MADX
lattice. You can create a lattice using one of the predefined simple 
lattices or by instantiating the Machine class and adding many 
elements to it using its various Add methods. This instance may be
written out to a MADX input text file using the WriteMachine method.

Classes:

Element - beam line element that always has name and type
Line    - a list of elements
Machine - a sequence of elements and associated options and beam etc.

"""

import _General
from   _General import IsFloat as _IsFloat
from   decimal import Decimal as _Decimal
import time

from pymadx.Beam import Beam as _Beam

madxcategories = [
    'drift',
    'solenoid',
    'sbend',
    'rbend',
    'hkicker',
    'vkicker',
    'quadrupole',
    'sextupole',
    'octupole',
    'decapole',
    'multipole',
    'marker'
    ]

class Element(dict) :
    """
    Element - a beam element class - inherits dict

    Element(name,type,**kwargs)

    A beam line element must ALWAYs have a name, and type.
    The keyword arguments are specific to the type and are up to
    the user to specify.

    Numbers are converted to a python Decimal type to provide
    higher accuracy in the representation of numbers - 15
    decimal places are used.
    """
    def __init__(self, name, category, **kwargs):
        if category not in madxcategories:
            raise ValueError("Not a valid MADX element type")
        self.name     = str(name)
        self.category = str(category)
        self.length      = 0.0 #for bookeeping only
        self['name']     = self.name
        self['category'] = self.category
        self._keysextra = []
        for key,value in kwargs.iteritems():
            if type(value) == tuple and category != 'multipole':
                #use a tuple for (value,units)
                self[key] = (_Decimal(str(value[0])),value[1])
            elif type(value) == tuple and category == 'multipole' :
                self[key] = value
            elif _IsFloat(value):
                #just a number
                self[key] = _Decimal(str(value))
            else:
                #must be a string
                self[key] = '"'+value+'"'
            self._keysextra.append(str(key)) #order preserving
        if 'l' in self:
            if type(self['l']) == tuple:
                ll = self['l'][0]
            else:
                ll = self['l']
            self.length += float(ll)

    def keysextra(self):
        #so behaviour is similar to dict.keys()
        return self._keysextra

    def __repr__(self):
        s = ''
        s += self.name + ': '
        s += self.category
        if len(self._keysextra) > 0:
            for key in self._keysextra:
                if type(self[key]) == tuple and self.category != 'multipole':
                    s += ', ' + key + '=' + str(self[key][0]) + '*' + str(self[key][1])
                elif type(self[key]) == tuple and self.category == 'multipole' :
                    s += ', ' + key + '=' + '{'+(','.join([str(s) for s in self[key]]))+'}'
                else:
                    s += ', ' + key + '=' + str(self[key])
        s += ';\n'
        return s

class Line(list):
    def __init__(self,name,*args):
        for item in args[0]:
            if type(item) != Element:
                raise TypeError("Line is a list of Elements")
        list.__init__(self,*args)
        self.name   = name
        self.length = 0.0 
        for item in args[0]:
            self.length += item.length

    def __repr__(self):
        s = ''
        for item in self:
            s += str(item)+'\n' #uses elements __repr__ function
        s += self.name+ ': line=('
        s += ', '.join([item.name for item in self]) + ');\n'
        return s

    def DefineConstituentElements(self):
        s = ''
        for item in self:
            s += str(item) #uses elements __repr__ function
        return s

class Sampler(object):
    """
    Class that can return the appropriate sampler syntax if required.
    """ 
    def __init__(self,name):
        self.name = name

    def __repr__(self):
        return 'PTC_OBSERVE, place='+self.name+';\n'

class Machine(object):     
    def __init__(self,verbose=False):
        self.verbose   = verbose
        self.sequence  = []
        self.elements  = []
        self.elementsd = {}
        self.samplers  = []
        self.length    = 0.0
        self.angint    = 0.0
        self.beam      = _Beam() # always construct at least default beam
        self.ptcaper   = []

    def __repr__(self):
        s = ''
        s += 'pymadx.Builder.Machine instance\n'
        s += str(len(self.sequence)) + ' items in sequence\n'
        s += str(len(self.elements)) + ' unique elements defined\n'
        return s

    def __iter__(self):
        self._iterindex = -1
        return self

    def next(self):
        if self._iterindex == len(self.sequence)-1:
            raise StopIteration
        self._iterindex += 1
        return self.elementsd[self.sequence[self._iterindex]]
        
    def __getitem__(self,name):
        if _IsFloat(name):
            return self.elementsd[self.sequence[name]]
        else:
            return self.elementsd[name]

    def __len__(self):
        return len(self.elementsd.keys())

    def Append(self,object):
        if type(object) not in (Element,Line):
            raise TypeError("Only Elements or Lines can be added to the machine")
        elif object.name not in self.sequence:
            #hasn't been used before - define it
            if type(object) is Line:
                for element in object:
                    self.elements.append(element)
                    self.elementsd[element.name] = element
                self.elements.append(object)
                self.elementsd[object.name] = object
            else:
                self.elements.append(object)
                self.elementsd[object.name] = object
        #finally add it to the sequence
        self.sequence.append(object.name)
        self.length += object.length

    def Write(self,filename,verbose=False):
        """
        Write the machine to a series of gmad files.
        """
        verboseresult = verbose or self.verbose
        WriteMachine(self,filename,verboseresult)

    def AddDrift(self, name='dr', length=0.1, **kwargs):
        if self.verbose:
            print 'AddDrift>  ',name,' ',length,' ',kwargs
        if length < 1e-12:
            self.AddMarker(name)
        else:
            self.Append(Element(name,'drift',l=length,**kwargs))

    def AddDipole(self, name='dp', category='sbend', length=0.1, angle=0.001, **kwargs):
        """
        AddDipole(category='sbend')

        category - 'sbend' or 'rbend' - sector or rectangular bend
        """
        if self.verbose:
            print 'AddDipole> ',name,' ',length,' ',kwargs
        if length < 1e-12:
            self.AddMarker(name)
        self.Append(Element(name,category,l=length,angle=angle,**kwargs))

    def AddHKicker(self, name='hk', hkick=0, length=0, **kwargs):
        self.Append(Element(name,'hkicker',hkick=hkick,l=length,**kwargs))

    def AddVKicker(self, name='vk', vkick=0, length=0, **kwargs):
        self.Append(Element(name,'vkicker',vkick=vkick,l=length,**kwargs))

    def AddQuadrupole(self, name='qd', length=0.1, k1=0.0, **kwargs):
        self.Append(Element(name,'quadrupole',l=length,k1=k1,**kwargs))

    def AddSextupole(self, name='sd', length=0.1, k2=0.0, **kwargs): 
        self.Append(Element(name,'sextupole',l=length,k2=k2,**kwargs))

    def AddOctupole(self, name='oc', length=0.1, k3=0.0, **kwargs): 
        self.Append(Element(name,'octupole',l=length,k3=k3,**kwargs))

    def AddDecapole(self, name='dd', length=0.1, k4=0.0, **kwargs): 
        self.Append(Element(name,'decapole',l=length,k4=k4,**kwargs))

    def AddMultipole(self, name='mp', knl=(0), ksl=(0), **kwargs): 
        self.Append(Element(name,'multipole', knl=knl, ksl=ksl, **kwargs))

    def AddSampler(self,*elementnames):
        if elementnames[0] == 'all':
            for element in self.elements:
                #remember we can only have samplers on uniquely
                #named elements (for now)
                self.samplers.append(Sampler(element.name))
        elif elementnames[0] == 'first':
            self.samplers.append(Sampler(self.elements[0].name))
        elif elementnames[0] == 'last':
            self.samplers.append(Sampler(self.elements[-1].name))
        else:
            for element in elementnames:
                if element not in self.sequence:
                    raise ValueError(element+" is not a valid element in this machine")
                else:
                    self.samplers.append(Sampler(element))
        
    def AddBeam(self, beam=None) : 
        self.beam = beam

    def AddMarker(self, name="mk", **kwargs):
        self.Append(Element(name,'marker', **kwargs))

    def AddSolenoid(self, name='sl', length=0.1, ks=0.0, **kwargs):
        self.Append(Element(name,'solenoid',l=length,ks=ks,**kwargs))

    def AddOptions(self,*args,**kwargs):
        pass

    def AddPTCTrackAperture(self, aperture=[]):
        """
        Add a PTC 6D max aperture for ptc_track command.
        """
        if len(aperture) > 0 and len(aperture) != 6:
            raise IndexError("Aperture list must contain exactly 6 values")
        # convert all to float to ensure it can be written
        aper = [float(i) for i in aperture]
        self.ptcaper = aper

# General scripts below this point

def WriteMachine(machine, filename, verbose=False):
    """
    WriteMachine(machine(machine),filename(string),verbose(bool))
    
    Write a lattice to disk. This writes several files to make the
    machine, namely:
    
     * filename_components.madx - component files (max 10k per file)
     * filename_sequence.madx   - lattice definition
     * filename_samplers.madx   - sampler definitions (max 10k per file)
     * filename.gmad            - suitable main file with all sub 
                                  files in correct order
    
    These are prefixed with the specified filename / path.
    """
    
    if not isinstance(machine,Machine):
        raise TypeError("Not machine instance")
    
    elementsperline = 100 #number of machine elements per madx line (not text line)
    
    #check filename
    if filename[-5:] != '.madx':
        filename += '.madx'
    #check if file already exists
    ofilename = filename
    filename = _General.CheckFileExists(filename)
    if filename != ofilename:
        print 'Warning, chosen filename already exists - using filename: ',filename.split('.')[0]
    basefilename = filename[:-5]#.split('/')[-1]

    #prepare names
    files         = []
    fn_main       = basefilename + '.madx'
    fn_components = basefilename + '_components.madx'
    fn_sequence   = basefilename + '_sequence.madx'
    fn_beam       = basefilename + '_beam.madx'
    fn_options    = basefilename + '_options.madx'
    fn_ptc        = basefilename + '_ptcjob.madx'
    timestring = '! ' + time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime()) + '\n'

    # write beam 
    f = open(fn_beam,'w') 
    files.append(fn_beam)
    f.write(timestring) 
    f.write('! pymadx.Builder \n')
    f.write('! BEAM DEFINITION \n\n')
    f.write(machine.beam.ReturnBeamString())
    f.close()
    
    #write component files
    f = open(fn_components,'w')
    files.append(fn_components)
    f.write(timestring)
    f.write('! pymadx.Builder Machine \n')
    f.write('! COMPONENT DEFINITION\n\n')
    for element in machine.elements:
        f.write(str(element))
    f.close()

    #write lattice sequence
    f = open(fn_sequence,'w')
    files.append(fn_sequence)
    f.write(timestring)
    f.write('! pymadx.Builder Machine \n')
    f.write('! LATTICE SEQUENCE DEFINITION\n\n')
    sequencechunks = _General.Chunks(machine.sequence,elementsperline)
    linelist = []
    ti = 0
    for line in _General.Chunks(machine.sequence,elementsperline):
        f.write('l'+str(ti)+': line = ('+', '.join(line)+');\n')
        linelist.append('l'+str(ti))
        ti += 1
    # need to define the period before making sampler planes
    f.write('lattice: line = ('+', '.join(linelist)+');\n')
    f.write('use, period=lattice;\n')
    f.close()

    # optionally write start of ptc job / inialise the universe
    if machine.beam['distrType'] == 'ptc':
        f = open(fn_ptc,'w')
        files.append(fn_ptc)
        f.write(timestring)
        f.write('! pymadx.Builder Machine \n')
        f.write('! PTC JOB SPECIFICATION\n\n')
        f.write(machine.beam.ReturnPtcString())

        ptcTrackstr = ''
        # check attribute existance in case an old machine instance is supplied
        if hasattr(machine, 'ptcaper'):
            # check for correct no. of aperture dimensions
            if len(machine.ptcaper) > 0 and len(machine.ptcaper) != 6:
                raise IndexError("Aperture list must contain exactly 6 values")
            # string with ptc track apertures if defined.
            if len(machine.ptcaper) > 0:
                ptcTrackstr = ', maxaper={'
                for index,aper in enumerate(machine.ptcaper):
                    ptcTrackstr += str(aper)
                    if index != (len(machine.ptcaper)-1):
                        ptcTrackstr += ', '
                ptcTrackstr += '}'

        #write samplers - only for PTC jobs
        if len(machine.samplers) > 0:
            f.write('\n! SAMPLER DEFINITION\n\n')
            for sampler in machine.samplers:
                f.write(str(sampler))
            f.write('ptc_track, element_by_element, dump, turns=1, icase=5, onetable' + ptcTrackstr + ';\n')
            f.write('PTC_TRACK_END;\n')
            f.write('ptc_end;\n')
        f.close()

    # write main file
    f = open(fn_main,'w')
    f.write(timestring)
    f.write('! pymadx.Builder Machine \n')
    f.write('! number of elements = ' + str(len(machine.elements)) + '\n')
    f.write('! total length       = ' + str(machine.length) + ' m\n\n')
    f.write('set, format="22.16e";\n')
    
    for fn in files:
        fn = fn.split('/')[-1]
        f.write("call, file='"+fn+"';\n")

    # line in main file for outputting twiss params to .tfs file
    if not machine.beam['distrType'] == 'ptc':
        f.write('\n')
        f.write(machine.beam.ReturnTwissString(basefilename))

    f.close()

    #user feedback
    print 'Machine written to:'
    for fn in files:
        print(fn)
    print 'All included in main file: \n',fn_main



"""
Conversion from and to various formats related to the MadX format

See individual methods for documentation

"""

from _Mad8ToMadx import Mad8ToMadx
from _TfsToPtc import TfsToPtc
from _Transport2Madx import Transport2Madx



# http://mad.web.cern.ch/mad/madx.old/mad8tomadX/
# constant 
# multipole

import os as _os 
import re as _re 

def ConvertDir(inputDir) : 
    files = _os.listdir(inputDir) 
    for f in files : 
        if f.find('~') == -1 : 
            Mad8ToMadX(inputDir+'/'+f)
        
def Mad8ToMadX(inputName) : 
    inputFile  = open(inputName) 
    outputName  = inputName[inputName.rfind('/')+1:inputName.rfind('.')]+".xsifx"
    outputFile = open(outputName,'w')
    
    print 'Mad8ToMadX, input > ',inputName
    print 'Mad8ToMadX, output> ',outputName


    ml = '' # merged line 
    inputFile1 = []
    for l in inputFile : 
        l  = l.rstrip() 
        
        # find line continuations (apersand index) 
        ai = l.find('&')
        if ai != -1 : 
            l = l.replace('&','')
            ml = ml+l
        else :
            if len(ml) == 0 :
                ml = l
            else :
                ml = ml+l
            inputFile1.append(ml)
            ml = ''
        
    ltc = ';'
    for l in inputFile1 : 
        l = l.rstrip()
        
        # find comments (comment index)
        ci = l.find('!')

        # Deal with each line
        if ci == -1 and len(l) == 0: # Empty line  
            pl = l
        elif ci == -1 and len(l) > 0 : # No comment and actual line 
            pl = l+ltc
        elif ci == 0 : # Whole line is comment
            pl = l 
        else : # comment present 
            if len(l) == 0 : 
                print ci
            # split line on comment 
            bcl = l[0:ci] # before comment
            acl = l[ci:]  # after comment
            pl  = bcl+ltc+acl
 
        #######################################################################################
        # Simple replacements 
        #######################################################################################
        # replace intr with intrument 
        if pl.find('INSTRUMENT') == -1 :
            pl = pl.replace('INST','INSTRUMENT')        
        # replace moni with monitor
        if pl.find('MONITOR') == -1 :
            pl = pl.replace('MONI','MONITOR')
        # replace mark with marker 
        if pl.find('MARKER') == -1 :
            pl = pl.replace('MARK','MARKER')
        # replace drif with drift 
        if pl.find('DRIFT') == -1 :
            pl = pl.replace('DRIF','DRIFT')
        # replace quad with quadrupole 
        if pl.find('QUADRUPOLE') == -1 :
            pl = pl.replace('QUAD','QUADRUPOLE') 
        # replace sext with sextupole 
        if pl.find('SEXTUPOLE') == -1 : 
            pl = pl.replace('SEXT','SEXTUPOLE')
        # replace octu with octupole
        if pl.find('OCTUPOLE') == -1 : 
            pl = pl.replace('OCTU','OCTUPOLE')
        # replace sben with sbend
        if pl.find('SBEND') == -1 :
            pl = pl.replace('SBEN','SBEND')
        # replace aper with aperture 
        if pl.find('APERTURE') == -1 :
            pl = pl.replace('APER','APERTURE') 
        # replce ecoll with ecollimator
        if pl.find('ECOLLIMATOR') == -1 :
            pl = pl.replace('ECOLL','ECOLLIMATOR') 
        if pl.find('ECOLLIMATOR') == -1 :
            pl = pl.replace('ECOL','ECOLLIMATOR') 
        # replce rcoll with ecollimator
        if pl.find('RCOLLIMATOR') == -1 :
            pl = pl.replace('RCOLL','RCOLLIMATOR') 
        if pl.find('RCOLLIMATOR') == -1 :
            pl = pl.replace('RCOL','RCOLLIMATOR')
        if pl.find('HKICKER') == -1 : 
            pl = pl.replace('HKICK','HKICKER')
        if pl.find('HKICKER') == -1 : 
            pl = pl.replace('HKIC','HKICKER')
        if pl.find('VKICKER') == -1 : 
            pl = pl.replace('VKICK','HKICKER')
        if pl.find('VKICKER') == -1 : 
            pl = pl.replace('VKIC','HKICKER')



        #######################################################################################
        # Regular expressions 
        #######################################################################################
        # constant 
        if pl.find('CONSTANT') != -1 : 
#            print pl
            m = _re.search('(\w+)\s*:\s*CONSTANT\s*=\s*([A-Za-z0-9+-/*.()]+)',pl)
            pl = '   const '+m.group(1)+'='+m.group(2)+';'
            print pl

        # multipole 
        if pl.find('MULTIPOLE') != -1: 
            print pl
            n  = _re.search('([A-Za-z0-9.]+)\s*:\s*MULTIPOLE',pl)
            m0 = _re.search('K0L\s*=\s*([A-Za-z0-9+-/*.()]+),',pl)
            m1 = _re.search('K1L\s*=\s*([A-Za-z0-9+-/*.()]+),',pl)
            m2 = _re.search('K2L\s*=\s*([A-Za-z0-9+-/*.()]+),',pl)
            m3 = _re.search('K3L\s*=\s*([A-Za-z0-9+-/*.()]+),',pl)
            m4 = _re.search('K4L\s*=\s*([A-Za-z0-9+-/*.()]+),',pl)
            m5 = _re.search('K5L\s*=\s*([A-Za-z0-9+-/*.()]+),',pl)
            m6 = _re.search('K6L\s*=\s*([A-Za-z0-9+-/*.()]+),',pl)
            mt0 = _re.search('T0\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)       # match t0
            mt1 = _re.search('T1\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)       # match t1
            mt2 = _re.search('T2\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)       # match t2
            mt3 = _re.search('T3\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)       # match t3
            mt4 = _re.search('T4\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)       # match t4
            mt5 = _re.search('T5\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)       # match t5
            mt6 = _re.search('T6\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)       # match t6
            ml  = _re.search(' L\s*=\s*([0-9.E+-])',pl)                    # match length

            mt  = _re.search('TYPE\s*=\s*("(.*?)")',pl)                    # match type
            mr  = _re.search('LRAD\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl)     # match lrad
            ma  = _re.search('APERTURE\s*=\s*([A-Za-z0-9+-/*.()]+)\s*',pl) # match aperture

            pl = n.group(1)+' : MULTIPOLE '

            if ml : 
                pl = pl+', L='+ml.group(1)

            k0l = '0.0'; t0 = '0.0'
            k1l = '0.0'; t1 = '0.0'
            k2l = '0.0'; t2 = '0.0'
            k3l = '0.0'; t3 = '0.0'
            k4l = '0.0'; t4 = '0.0'
            k5l = '0.0'; t5 = '0.0'
            k6l = '0.0'; t6 = '0.0'
                
            if m0 : 
                k0l = m0.group(1)
            if m1 : 
                k1l = m1.group(1)
            if m2 : 
                k2l = m2.group(1)
            if m3 : 
                k3l = m3.group(1)
            if m4 : 
                k4l = m4.group(1)
            if m5 : 
                k5l = m5.group(1)
            if m6 : 
                k6l = m6.group(1)

            if mt0 : 
                t0 = mt0.group(1)
            if mt1 : 
                t1 = mt1.group(1)
            if mt2 : 
                t2 = mt2.group(1)
            if mt3 : 
                t3 = mt3.group(1)
            if mt4 : 
                t4 = mt4.group(1)
            if mt5 : 
                t5 = mt5.group(1)
            if mt6 : 
                t6 = mt6.group(1)            

            pl = pl+ ', KNL={'+k0l+','+k1l+','+k2l+','+k3l+','+k4l+','+k5l+','+k6l+'}'
            
            if mt : 
                pl = pl +', TYPE='+mt.group(1)
            if ma : 
                pl = pl +', APERTURE='+ma.group(1)
            if mr : 
                pl = pl +', LRAD='+mr.group(1)

            pl = pl+';'

            print pl
                
        # attributes 
        def maFunc(m) : 
            return m.group(1)+'->'+m.group(2)

        pl = _re.sub('([a-zA-Z0-9_-]+)\[([a-zA-Z]+)\]',maFunc,pl) # replace [] with ->
        outputFile.write(pl+'\n')



"""
Class to create a MADX beam definition.
"""

MADXDistributionTypes = [
    'reference',
    'madx',
    'ptc'
]

MADXParticleTypes = [
    'electron',
    'positron',
    'proton'
]

class Beam(dict):
    """
    A class that extends a dictionary for the specific parameters
    in a MADX beam definition. This class can return a string 
    representation of itself that is valid MADX syntax.

    Setter methods are dynamically added based on the distribution selected.
    """
    def __init__(self,particletype='e-',energy=1.0,distrtype='reference',*args,**kwargs):
        dict.__init__(self,*args,**kwargs)
        self.isPTCDistribution = False
        self.SetParticleType(particletype)
        self.SetEnergy(energy)
        self.SetDistributionType(distrtype)

    def __repr__(self):
        return self.ReturnBeamString()

    def GetItemStr(self, key):
        return str(self[key])
        
    def SetParticleType(self,particletype='e-'):
        if particletype == 'e-' : 
            particletype = 'electron' 
        elif particletype == 'e+' : 
            particletype = 'positron' 
        elif particletype == 'proton' :
            particletype = 'proton'

        if particletype not in MADXParticleTypes:
            raise ValueError("Unknown particle type: '"+str(particletype)+"'")
        self['particle'] = str(particletype) 

    def SetEnergy(self,energy=1.0,unitsstring='GeV'):
        self['energy'] = str(energy)

    def SetDistributionType(self,distrtype='reference'):
        if distrtype not in MADXDistributionTypes:
            raise ValueError("Unknown distribution type: '"+str(distrtype)+"'")
        
        self['distrType'] = distrtype 
        if distrtype == 'madx':
            setattr(self, 'SetBetaX',            self._SetBetaX)
            setattr(self, 'SetBetaY',            self._SetBetaY) 
            setattr(self, 'SetAlphaX',           self._SetAlphaX)
            setattr(self, 'SetAlphaY',           self._SetAlphaY)
            setattr(self, 'SetEmittanceX',       self._SetEmittanceX) 
            setattr(self, 'SetEmittanceY',       self._SetEmittanceY)
            setattr(self, 'SetSigmaE',           self._SetSigmaE)
            setattr(self, 'SetSigmaT',           self._SetSigmaT)
        elif distrtype == 'ptc':
            self.isPTCDistribution = True
            setattr(self, 'SetDistribFileName',  self._SetDistribFileName)

    def WriteToFile(self, filename):
        f = open(filename, 'w')
        f.write(self.ReturnBeamString())
        f.close()
        
    def ReturnBeamString(self):
        s = 'beam, particle = {}, energy = {}'.format(self['particle'],
                                                      self['energy'])
        try:
            s += ', ex={}, ey={}'.format(self['emitx'], self['emity'])
        except KeyError:
            pass
        try:
            s += ', sige = {};\n'.format(self['sigmaE'])
        except KeyError:
            s += ';\n'
        return s

    def ReturnPtcString(self) : 
        s = 'ptc_create_universe;\n' 
        s+= 'ptc_create_layout,model=2,method=6,nst=10,time=false;\n'
        s+= 'call, file ="'+self['distrFile']+'";\n'
        s+= 'ptc_align;'
        return s
    
    def ReturnTwissString(self,basefilename='output'):
        s = 'twiss, save, '
        s += 'betx=' + self['betx'] + ', bety=' + self['bety']
        s += ', file=' + basefilename+'.tfs;'
        return s

    def SetX0(self,x0=0.0):
        self['X0'] = x0

    def SetY0(self,y0=0.0):
        self['Y0'] = y0
    
    def SetXP0(self,xp0=0.0):
        self['Xp0'] = xp0

    def SetYP0(self,yp0=0.0):
        self['Yp0'] = yp0

    def SetT0(self,t0=0.0,unitsstring='s'):
        self['T0'] = t0 + '*' + unitsstring

    def _SetSigmaE(self,sigmae=0.001):
        """
        fractional energy spread
        """
        self['sigmaE'] = sigmae

    def _SetSigmaT(self,sigmat=1.0,unitsstring='um'):
        self['sigmaT'] = sigmat

    def _SetBetaX(self,betx=1.0,unitsstring='m'):
        self['betx'] = str(betx)# + '*' + unitsstring

    def _SetBetaY(self,bety=1.0,unitsstring='m'):
        self['bety'] = str(bety)# + '*' + unitsstring

    def _SetAlphaX(self,alphax=1.0,unitsstring='m'):
        self['alfx'] = str(alphax)# + '*' + unitsstring

    def _SetAlphaY(self,alphay=1.0,unitsstring='m'):
        self['alfy'] = str(alphay)# + '*' + unitsstring

    def _SetEmittanceX(self,emitx=1.0,unitsstring='um'):
        self['emitx'] = str(emitx)# + '*' + unitsstring
   
    def _SetEmittanceY(self,emity=1.0,unitsstring='um'):
        self['emity'] = str(emity)# + '*' + unitsstring

    def _SetDistribFileName(self, fileName) :
        self['distrFile'] = fileName



import pytest

import pymadx

@pytest.fixture()
def reference_beam():
    return pymadx.Beam.Beam(particletype='e-',
                            energy=1.0,
                            distrtype='reference')

def test_SetEnergy(reference_beam):
    reference_beam.SetEnergy(10.0)
    assert reference_beam.GetItemStr('energy') == '10.0'

@pytest.mark.parametrize('test_input, expected',
                         [('e-', 'electron'),
                          ('e+', 'positron'),
                          ('proton', 'proton')])
def test_SetParticleType(reference_beam, test_input, expected):
    reference_beam.SetParticleType(test_input)
    assert reference_beam['particle'] == expected

def test_SetParticleType_raises_exception(reference_beam):
    with pytest.raises(ValueError):
        reference_beam.SetParticleType('nonsense')



"""
Classes to handle PTC runs and data.
"""

import numpy as _np
import re as _re
try:
    import matplotlib.pyplot as _plt
except ImportError:
    pass
from numpy.random import multivariate_normal as _multivariate_normal

class Inray(object):
    """
    Class for a madx ptc input ray
    x  : horizontal position [m]
    px : horizontal canonical momentum p_x/p_0 
    y  : vertical position [m]
    py : vertical canonical momentum p_y/p_0
    t  : c*(t-t0) [m]
    pt : (delta-E)/(pc)

    use str(Inray) to get the representation for file writing
    """
    def __init__(self, x=0.0, px=0.0, y=0.0, py=0.0, t=0.0, pt=0.0):
        self.x  = x
        self.px = px
        self.y  = y
        self.py = py
        self.t  = t
        self.pt = pt

    def __repr__(self):
        s =  'ptc_start'
        s += ', x='  + str(self.x)
        s += ', px=' + str(self.px)
        s += ', y='  + str(self.y)
        s += ', py=' + str(self.py)
        s += ', t='  + str(self.t)
        s += ', pt=' + str(self.pt)
        s += ';\n'
        return s

class Inrays(list):
    """
    Class based on python list for Inray class
    """
    def __init__(self):
        list.__init__(self)
        variables = ['X','PX','Y','PY','T','PT']
        for v in variables:
            self._AddMethod(v)

    def AddParticle(self,x=0.0,px=0.0,y=0.0,py=0.0,t=0.0,pt=0.0):
        self.append(Inray(x,px,y,py,t,pt))

    def Clear(self):
        del self[:]

    def Write(self,filename):
        WriteInrays(filename,self)

    def Plot(self):
        PlotInrays(self)

    def _AddMethod(self, variablename):
        """This is used to easily and dynamically add a getter function for a variable name."""
        def GetAttribute():
            return _np.array([getattr(p,str.lower(variablename)) for p in self])
        setattr(self,variablename,GetAttribute)

    def Statistics(self):
        print 'TBC - will return various moments'
        
def LoadInrays(fileName): 
    """Load input rays from file
    fileName : inrays.madx 
    return   : Inrays instance""" 
    i = Inrays()
    
    # open file 
    f = open(fileName) 
    for l in f : 
        inre_x  = _re.search('\s*x\s*=\s*([0-9.eE+-]+)\s*',l)
        inre_px = _re.search('\s*px\s*=\s*([0-9.eE+-]+)\s*',l)
        inre_y  = _re.search('\s*y\s*=\s*([0-9.eE+-]+)\s*',l)
        inre_py = _re.search('\s*py\s*=\s*([0-9.eE+-]+)\s*',l)
        inre_t  = _re.search(' t\s*=\s*([0-9.eE+-]+)\s*',l)
        inre_pt = _re.search('\s*pt\s*=\s*([0-9.eE+-]+)\s*',l)

        if inre_x : 
            x = float(inre_x.group(1))
        else :
            x = 0.0
        if inre_px : 
            px = float(inre_px.group(1))
        else : 
            px = 0.0 
        if inre_y : 
            y  = float(inre_y.group(1))
        else : 
            y  = 0.0
        if inre_py :             
            py = float(inre_py.group(1))
        else : 
            py = 0.0
        if inre_t :             
            t  = float(inre_t.group(1))
        else : 
            t = 0.0

        if inre_pt:
            pt = float(inre_pt.group(1))
        else:
            pt = 0.0

        i.AddParticle(x,px,y,py,t,pt)

    print 'LoadInrays> Loaded ',len(i)
    return i
  
def WriteInrays(fileName, inrays):
    print 'pymadx.Ptc> WriteInrays - inrays written to: ',fileName
    f = open(fileName, 'w') 
    for particle in inrays:
        f.write(str(particle))
    f.close()
   
def PlotInrays(i): 
    """Plot Inrays instance, if input is a sting the instance is created from the file"""    

    if type(i) == str : 
        i = LoadInrays(i)
        
    f = _plt.figure(1) 
    f.clf()
    
    ax = f.add_subplot(3,2,1)    
    ax.hist(i.X(),50,histtype='step',label='x')
    ax.text(0.05,0.85,'X',transform=ax.transAxes)
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    ax = f.add_subplot(3,2,2)    
    ax.hist(i.PX(),50,histtype='step',label='px')
    ax.text(0.05,0.85,'PX',transform=ax.transAxes)
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    
    ax = f.add_subplot(3,2,3)    
    ax.hist(i.Y(),50,histtype='step',label='y')
    ax.text(0.05,0.85,'Y',transform=ax.transAxes)
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    ax = f.add_subplot(3,2,4)    
    ax.hist(i.PY(),50,histtype='step',label='py')
    ax.text(0.05,0.85,'PY',transform=ax.transAxes)
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    
    ax = f.add_subplot(3,2,5)    
    ax.hist(i.T(),50,histtype='step',label='t')
    ax.text(0.05,0.85,'T',transform=ax.transAxes)
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    ax = f.add_subplot(3,2,6)    
    ax.hist(i.PT(),50,histtype='step',label='pt')
    ax.text(0.05,0.85,'PT',transform=ax.transAxes)
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
    
    _plt.subplots_adjust(hspace=0.35,wspace=0.15,top=0.98,right=0.98,left=0.05)
  
class GaussGenerator(object): 
    """Simple ptx inray file generator"""
    def __init__(self,
                 gemx = 1e-10, betax = 0.1, alfx = 0.0 , 
                 gemy = 1e-10, betay = 0.1, alfy = 0.0,
                 sigmat = 1e-12, sigmapt= 1e-12): 
        """Simple gaussian beam
        gemx   : x geometric emittance 
        betax  : x beta function
        alfx   : x alpha function
        gemy   : y geometric emittance
        betay  : y beta function
        alfy   : y alpha function 
        sigmat : gaussian spread in time around the reference time 
        sigmapt: gaussian spread in relative energy 
        """
        
        self.gemx    = gemx
        self.betax   = betax
        self.alfx    = alfx

        self.gemy    = gemy
        self.betay   = betay
        self.alfy    = alfy 

        self.sigmat  = sigmat 
        self.sigmapt = sigmapt

        self.gamx    = (1.0+self.alfx**2)/self.betax    
        self.gamy    = (1.0+self.alfy**2)/self.betay

        # Generate sigma matrix
        self.means  = _np.zeros(6)
        self.sigmas = _np.zeros((6,6)) 
        
        self.sigmas[0][0] =  self.gemx*self.betax
        self.sigmas[0][1] = -self.gemx*self.alfx
        self.sigmas[1][0] = -self.gemx*self.alfx
        self.sigmas[1][1] =  self.gemx*self.gamx
        self.sigmas[2][2] =  self.gemy*self.betay
        self.sigmas[2][3] = -self.gemy*self.alfy
        self.sigmas[3][2] = -self.gemy*self.alfy
        self.sigmas[3][3] =  self.gemy*self.gamy
        self.sigmas[4][4] =  self.sigmat**2
        self.sigmas[5][5] =  self.sigmapt**2
        
    def __repr__(self) : 
        s = 'ex : '+str(self.gemx)+' bx : '+str(self.betax)+' ax : '+str(self.alfx)+' gx : '+str(self.gamx)+'\n'
        s+= 'ey : '+str(self.gemy)+' bx : '+str(self.betay)+' ax : '+str(self.alfy)+' gy : '+str(self.gamy)+'\n'
        s+= 'sT : '+str(self.sigmat)+' spt : '+str(self.sigmapt)
        return s

    def Generate(self, nToGenerate=1000, fileName='inrays.madx'): 
        """ returns an Inrays structure""" 
        i = Inrays()
        
        for c in range(0,nToGenerate,1):
            rv = _multivariate_normal(self.means,self.sigmas,1)[0]
            i.AddParticle(rv[0],rv[1],rv[2],rv[3],rv[4],rv[5]) 

        WriteInrays(fileName,i)
        #return i

class FlatGenerator(object):
    """Simple ptc inray file generator - even distribution"""
    def __init__(self,
                 mux =0.0, widthx =1e-3,
                 mupx=0.0, widthpx=1e-3,
                 muy =0.0, widthy =1e-3,
                 mupy=0.0, widthpy=1e-3):
        self.mux     = mux
        self.muy     = muy
        self.widthx  = widthx
        self.widthy  = widthy
        self.mupx    = mupx
        self.mupy    = mupy
        self.widthpx = widthpx
        self.widthpy = widthpy

    def __repr__(self):
        s = '' #TBC
        return s

    def Generate(self,nToGenerate=100, fileName='inrays.madx'):
        """ returns an Inrays structure""" 
        i = Inrays()

        nd = 0.0
        if self.widthx > 0:
            nd +=1
        if self.widthy > 0:
            nd +=1
        if self.widthpx > 0:
            nd +=1
        if self.widthpy > 0:
            nd +=1
        nperdim = _np.ceil(nToGenerate**(1/nd))
        print "FlatGenerator> making array square - there'll be ",nperdim**nd,'particles'

        xmin = self.mux - self.widthx/2.0
        xmax = self.mux + self.widthx/2.0
        ymin = self.muy - self.widthy/2.0
        ymax = self.muy + self.widthy/2.0
        pxmin = self.mupx - self.widthpx/2.0
        pxmax = self.mupx + self.widthpx/2.0
        pymin = self.mupy - self.widthpy/2.0
        pymax = self.mupy + self.widthpy/2.0

        def GetX():
            return _np.linspace(ymin,ymax,nperdim)

        def GetY():
            return _np.linspace(xmin,xmax,nperdim)

        def GetXp():
            return _np.linspace(pymin,pymax,nperdim)

        def GetYp():
            return _np.linspace(pxmin,pxmax,nperdim)

        #we don't actually use 'd' here but just take advantage of list comprehensions
        d = [i.AddParticle(x=x,px=px,y=y,py=py) for x in GetX() for px in GetXp() for y in GetY() for py in GetYp()]

        WriteInrays(fileName,i)
        #return i



"""
Ploting script for madx TFS files using the pymadx Tfs class

"""

import numpy as _np
import matplotlib         as _matplotlib
import matplotlib.patches as _patches
import matplotlib.pyplot  as _plt

class _My_Axes(_matplotlib.axes.Axes):
    """
    Inherit matplotlib.axes.Axes but override pan action for mouse.
    Only allow horizontal panning - useful for lattice axes.
    """
    name = "_My_Axes"
    def drag_pan(self, button, key, x, y):
        _matplotlib.axes.Axes.drag_pan(self, button, 'x', x, y) # pretend key=='x'

#register the new class of axes
_matplotlib.projections.register_projection(_My_Axes)

def _GetOpticalDataFromTfs(tfsobject):
    """
    Utility to pull out the relevant optical functions into a simple dictionary.
    """
    d = {}
    d['s']         = tfsobject.GetColumn('S')
    d['betx']      = tfsobject.GetColumn('BETX')
    d['bety']      = tfsobject.GetColumn('BETY')
    d['dispx']     = tfsobject.GetColumn('DX')
    d['dispxbeta'] = tfsobject.GetColumn('DXBETA')
    #d['dispy']    = tfsobject.GetColumn('DY') #don't use
    d['x']         = tfsobject.GetColumn('X')
    d['y']         = tfsobject.GetColumn('Y')
    d['sigmax']    = tfsobject.GetColumn('SIGMAX')
    d['sigmay']    = tfsobject.GetColumn('SIGMAY')
    d['sigmaxp']   = tfsobject.GetColumn('SIGMAXP')
    d['sigmayp']   = tfsobject.GetColumn('SIGMAYP')
    return d

def Centroids(tfsfile, title='', outputfilename=None, machine=True):
    """
    Plot the centroid (mean) x and y from the a Tfs file or pymadx.Tfs instance.

    tfsfile        - can be either a string or a pymadx.Tfs instance.
    title          - optional title for plot
    outputfilename - optional name to save file to (extension determines format)
    machine        - if True (default) add machine diagram to top of plot
    """
    import pymadx.Data as _Data
    madx = _Data.CheckItsTfs(tfsfile)
    d    = _GetOpticalDataFromTfs(madx)
    smax = madx.smax

    f    = _plt.figure(figsize=(9,5))
    axoptics = f.add_subplot(111)

    #optics plots
    axoptics.plot(d['s'],d['x'],'b-', label=r'$\mu_{x}$')
    axoptics.plot(d['s'],d['y'],'g-', label=r'$\mu_{y}$')
    axoptics.set_xlabel('S (m)')
    axoptics.set_ylabel(r'$\mu_{(x,y)}$ (m)')
    axoptics.legend(loc=0,fontsize='small') #best position

    #add lattice to plot
    if machine:
        AddMachineLatticeToFigure(f,madx)

    _plt.suptitle(title,size='x-large')

    if outputfilename != None:
        if '.' in outputfilename:
            outputfilename = outputfilename.split('.')[0]
        _plt.savefig(outputfilename+'.pdf')
        _plt.savefig(outputfilename+'.png')

def Survey(tfsfile, title='', outputfilename=None):
    """
    Plot the x and z coordinates from a tfs file.
    """
    import pymadx.Data as _Data
    madx = _Data.CheckItsTfs(tfsfile)
    x    = madx.GetColumn('X')
    z    = madx.GetColumn('Z')

    f = _plt.figure()
    ax = f.add_subplot(111)
    ax.set_aspect('equal')

    ax.plot(x, z, marker='.')
    _plt.suptitle(title,size='x-large')
    _plt.xlabel('X (m)')
    _plt.ylabel('Z (m)')


def Beta(tfsfile, title='', outputfilename=None, machine=True, dispersion=False, squareroot=True):
    """
    Plot sqrt(beta x,y) as a function of S. By default, a machine diagram is shown at
    the top of the plot.

    Optionally set dispersion=True to plot x dispersion as second axis.
    Optionally turn off machine overlay at top with machine=False
    Specify outputfilename (without extension) to save the plot as both pdf and png.
    """
    import pymadx.Data as _Data
    madx = _Data.CheckItsTfs(tfsfile)
    d    = _GetOpticalDataFromTfs(madx)
    smax = madx.smax

    f    = _plt.figure(figsize=(9,5))
    axoptics = f.add_subplot(111)

    #optics plots
    if squareroot:
        yx = _np.sqrt(d['betx'])
        yy = _np.sqrt(d['bety'])
    else:
        yx = d['betx']
        yy = d['bety']
    axoptics.plot(d['s'], yx, 'b-', label='x')
    axoptics.plot(d['s'], yy, 'g-', label='y')
    if dispersion:
        axoptics.plot([], [],'r--', label=r'$\mathrm{D}_{x} / \beta (S)$') #fake plot for legend
    axoptics.set_xlabel('S (m)')
    if squareroot:
        axoptics.set_ylabel(r'$\sqrt{\beta}$ ($\sqrt{\mathrm{m}}$)')
    else:
        axoptics.set_ylabel(r'$\beta$ (m)')
    axoptics.legend(loc=0,fontsize='small') #best position

    #plot dispersion - only in horizontal
    if dispersion:
        ax2 = axoptics.twinx()
        ax2.plot(d['s'],d['dispxbeta'],'r--')
        ax2.set_ylabel(r'Dispersion / $\beta$ (m)')

    #add lattice to plot
    if machine:
        AddMachineLatticeToFigure(f,madx)

    _plt.suptitle(title,size='x-large')
    _plt.xlim((0 - 0.05*smax, 1.05*smax))
    if outputfilename != None:
        if '.' in outputfilename:
            outputfilename = outputfilename.split('.')[0]
        _plt.savefig(outputfilename+'.pdf')
        _plt.savefig(outputfilename+'.png')

def Sigma(tfsfile, title='', outputfilename=None, machine=True, dispersion=False):
    """
    Plot sqrt(beta x,y) as a function of S. By default, a machine diagram is shown at
    the top of the plot.

    Optionally set dispersion=True to plot x dispersion as second axis.
    Optionally turn off machine overlay at top with machine=False
    Specify outputfilename (without extension) to save the plot as both pdf and png.
    """
    import pymadx.Data as _Data
    madx = _Data.CheckItsTfs(tfsfile)
    d    = _GetOpticalDataFromTfs(madx)
    smax = madx.smax

    f    = _plt.figure(figsize=(9,5))
    axoptics = f.add_subplot(111)
    
    yx = d['sigmax']
    yy = d['sigmay']
    axoptics.plot(d['s'], yx, 'b-', label='x')
    axoptics.plot(d['s'], yy, 'g-', label='y')
    if dispersion:
        axoptics.plot([], [],'r--', label=r'$\mathrm{D}_{x} / \beta (S)$') #fake plot for legend
    axoptics.set_xlabel('S (m)')
    axoptics.set_ylabel(r'$\sigma$ (m)')
    axoptics.legend(loc=0,fontsize='small') #best position

    #plot dispersion - only in horizontal
    if dispersion:
        ax2 = axoptics.twinx()
        ax2.plot(d['s'],d['dispx'],'r--')
        ax2.set_ylabel(r'Dispersion / $\beta$ (m)')

    #add lattice to plot
    if machine:
        AddMachineLatticeToFigure(f,madx)

    _plt.suptitle(title,size='x-large')
    _plt.xlim((0 - 0.05*smax, 1.05*smax))
    if outputfilename != None:
        if '.' in outputfilename:
            outputfilename = outputfilename.split('.')[0]
        _plt.savefig(outputfilename+'.pdf')
        _plt.savefig(outputfilename+'.png')

def Aperture(aperture, machine=None, outputfilename=None, plot="xy", plotapertype=True):
    """
    Plots the aperture extents vs. S from a pymadx.Data.Aperture instance.

    Inputs:
      aperture (pymadx.Data.Aperture) - the aperture model to plot from
      machine (str or pymadx.Data.Tfs) - TFS file or TFS istance to plot a machine lattice from (default: None)
      outputfilename (str) - Name without extension of the output file if desired (default: None)
      plot (str) - Indicates whcih aperture to plot - 'x' for X, 'y' for Y and 'xy' for both (default: 'xy')
      plotapertype (bool) - If enabled plots the aperture type at every definted aperture point as a color-coded dot (default: False)
    """
    import pymadx.Data as _Data
    aper = _Data.CheckItsTfsAperture(aperture)

    allowed = ["x", "y", "xy", "X", "Y", "XY"]
    if plot not in allowed:
        raise ValueError("Invalid option plot: "+plot+". Use 'x', 'y' or 'xy'")

    f = _plt.figure(figsize=(9,5))

    s = aper.GetColumn('S')
    x,y = aper.GetExtentAll()

    if plotapertype:
        t = aper.GetColumn('APERTYPE')
        c = map(_ApertypeToColor, t)

    if "x" in plot.lower():
        line1, = _plt.plot(s, x, 'b-', label='X', alpha=0.6)
        if plotapertype:
            _plt.scatter(s, x, c=c, s=6)

    if "y" in plot.lower():
        line2, = _plt.plot(s, y, 'g-', label='Y', alpha=0.6)
        if plotapertype:
            _plt.scatter(s, y, c=c, s=6)

    _plt.xlabel('S (m)')
    _plt.ylabel('Aperture (m)')

    if plotapertype:
        _AddColorLegend(c)

    _plt.legend(loc='best', numpoints=1, scatterpoints=1, fontsize='small')

    if machine != None:
        AddMachineLatticeToFigure(_plt.gcf(), machine)

    if outputfilename != None:
        if '.' in outputfilename:
            outputfilename = outputfilename.split('.')[0]
        _plt.savefig(outputfilename+'.pdf')
        _plt.savefig(outputfilename+'.png')

def ApertureN1(aperture, machine=None, outputfilename=None):
    """
    Plot the N1 aperture value from MADX. 

    Requires N1 and S column.

    Optional "machine" argument is string to or pymadx.Data.Tfs instance
    for twiss description to provide a machine diagram on top.
    """
    
    import pymadx.Data as _Data
    aper = _Data.CheckItsTfsAperture(aperture)

    f = _plt.figure(figsize=(9,5))

    s = aper.GetColumn('S')
    n = aper.GetColumn('N1')

    _plt.plot(s,n)
    _plt.xlabel('S (m)')
    _plt.ylabel(r'N1 ($\sigma$)')

    if machine != None:
        AddMachineLatticeToFigure(_plt.gcf(), machine)

    if outputfilename != None:
        if '.' in outputfilename:
            outputfilename = outputfilename.split('.')[0]
        _plt.savefig(outputfilename+'.pdf')
        _plt.savefig(outputfilename+'.png')
        
def _ApertypeColorMap():
    #Some nice colors
    _colourCodes = ['#C03028',
                    '#F8D030',
                    '#6890F0',
                    '#F85888',
                    '#A8B820',
                    '#F08030',
                    '#7038F8',
                    '#78C850',
                    '#A8A878']

    #_colourCodes = [_HexToRGB(c) for c in _colourCodes]

    # MADX aperture types
    _madxAperTypes = ['CIRCLE',
                      'RECTANGLE',
                      'ELLIPSE',
                      'RECTCIRCLE',
                      'LHCSCREEN',
                      'MARGUERITE',
                      'RECTELLIPSE',
                      'RACETRACK',
                      'OCTAGON']
    typeToCol = dict(zip(_madxAperTypes, _colourCodes))
    return typeToCol

def _HexToRGB(h):
    h = h.strip('#')
    return tuple(int(h[i:i+2], 16) for i in (0, 2 ,4))

def _ApertypeToColor(apertype, cmap=_ApertypeColorMap()):
    color = (0,0,0)
    try:
        color = cmap[apertype.upper()]
    except:
        color =(0.8,0.8,0.8) # greyish

    return color

def _AddColorLegend(colors, cmap=_ApertypeColorMap()):
    found_cols = set(colors)
    typemap = dict((v,k) for k,v in cmap.iteritems()) #invert to get apertype from color
    handles = []
    for col in found_cols:
        _plt.scatter(None,None,color=col, label=typemap[col].lower())


def _SetMachineAxesStyle(ax):
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)

def _PrepareMachineAxes(figure):
    # create new machine axis with proportions 6 : 1
    axmachine = figure.add_subplot(911, projection="_My_Axes")
    axmachine.set_facecolor('none') # make background transparent to allow scientific notation
    _SetMachineAxesStyle(axmachine)
    return axmachine

def _AdjustExistingAxes(figure, fraction=0.9, tightLayout=True):
    """
    Fraction is fraction of height all subplots will be after adjustment.
    Default is 0.9 for 90% of height. 
    """
    # we have to set tight layout before adjustment otherwise if called
    # later it will cause an overlap with the machine diagram
    if (tightLayout):
        _plt.tight_layout()

    axs = figure.get_axes()

    for ax in axs:
        bbox = ax.get_position()
        bbox.y0 = bbox.y0 * fraction
        bbox.y1 = bbox.y1 * fraction
        ax.set_position(bbox)

def AddMachineLatticeToFigure(figure, tfsfile, tightLayout=True):
    """
    Add a diagram above the current graph in the figure that represents the
    accelerator based on a madx twiss file in tfs format.

    Note you can use matplotlib's gcf() 'get current figure' as an argument.

    >>> pymadx.Plot.AddMachineLatticeToFigure(gcf(), 'afile.tfs')

    A pymadx.Tfs class instance or a string specifying a tfs file can be
    supplied as the second argument interchangeably.

    """
    import pymadx.Data as _Data
    tfs = _Data.CheckItsTfs(tfsfile) #load the machine description

    #check required keys
    requiredKeys = ['KEYWORD', 'S', 'L', 'K1L']
    okToProceed = all([key in tfs.columns for key in requiredKeys])
    if not okToProceed:
        print "The required columns aren't present in this tfs file"
        print "The required columns are: ", requiredKeys
        raise IOError

    axoptics  = figure.get_axes()[0]
    _AdjustExistingAxes(figure, tightLayout=tightLayout)
    axmachine = _PrepareMachineAxes(figure)

    _DrawMachineLattice(axmachine,tfs)

    #put callbacks for linked scrolling
    def MachineXlim(ax):
        axmachine.set_autoscale_on(False)
        axoptics.set_xlim(axmachine.get_xlim())

    def Click(a) :
        if a.button == 3:
            try:
                print 'Closest element: ',tfs.NameFromNearestS(a.xdata)
            except ValueError:
                pass # don't complain if the S is out of bounds

    MachineXlim(axmachine)
    axmachine.callbacks.connect('xlim_changed', MachineXlim)
    figure.canvas.mpl_connect('button_press_event', Click)

def _DrawMachineLattice(axesinstance,pymadxtfsobject):
    ax  = axesinstance #handy shortcut
    tfs = pymadxtfsobject

    #NOTE madx defines S as the end of the element by default
    #define temporary functions to draw individual objects
    def DrawBend(e,color='b',alpha=1.0):
        br = _patches.Rectangle((e['S']-e['L'],-0.1),e['L'],0.2,color=color,alpha=alpha)
        ax.add_patch(br)
    def DrawQuad(e,color='r',alpha=1.0):
        if e['K1L'] > 0 :
            qr = _patches.Rectangle((e['S']-e['L'],0),e['L'],0.2,color=color,alpha=alpha)
        elif e['K1L'] < 0:
            qr = _patches.Rectangle((e['S']-e['L'],-0.2),e['L'],0.2,color=color,alpha=alpha)
        else:
            #quadrupole off
            qr = _patches.Rectangle((e['S']-e['L'],-0.1),e['L'],0.2,color='#B2B2B2',alpha=0.5) #a nice grey in hex
        ax.add_patch(qr)
    def DrawHex(e,color,alpha=1.0):
        s = e['S']-e['L']
        l = e['L']
        edges = _np.array([[s,-0.1],[s,0.1],[s+l/2.,0.13],[s+l,0.1],[s+l,-0.1],[s+l/2.,-0.13]])
        sr = _patches.Polygon(edges,color=color,fill=True,alpha=alpha)
        ax.add_patch(sr)
    def DrawRect(e,color,alpha=1.0):
        rect = _patches.Rectangle((e['S']-e['L'],-0.1),e['L'],0.2,color=color,alpha=alpha)
        ax.add_patch(rect)
    def DrawLine(e,color,alpha=1.0):
        ax.plot([e['S']-e['L'],e['S']-e['L']],[-0.2,0.2],'-',color=color,alpha=alpha)

    # plot beam line - make extra long in case of reversal - won't
    ax.plot([tfs.smin,tfs.smax],[0,0],'k-',lw=1)
    ax.set_ylim(-0.2,0.2)

    # loop over elements and Draw on beamline
    for element in tfs:
        kw = element['KEYWORD']
        if kw == 'QUADRUPOLE':
            DrawQuad(element, u'#d10000') #red
        elif kw == 'RBEND':
            DrawBend(element, u'#0066cc') #blue
        elif kw == 'SBEND':
            DrawBend(element, u'#0066cc') #blue
        elif kw == 'HKICKER':
            DrawRect(element, u'#4c33b2') #purple
        elif kw == 'VKICKER':
            DrawRect(element, u'#ba55d3') #medium orchid
        elif kw == 'RCOLLIMATOR':
            DrawRect(element,'k')
        elif kw == 'ECOLLIMATOR':
            DrawRect(element,'k')
        elif kw == 'COLLIMATOR':
            DrawRect(element,'k')
        elif kw == 'SEXTUPOLE':
            DrawHex(element, u'#ffcc00') #yellow
        elif kw == 'OCTUPOLE':
            DrawHex(element, u'#00994c') #green
        elif kw == 'DRIFT':
            pass
        elif kw == 'MULTIPOLE':
            DrawHex(element,'grey',alpha=0.5)
        else:
            #unknown so make light in alpha
            if element['L'] > 1e-1:
                DrawRect(element,'#cccccc',alpha=0.1) #light grey
            else:
                #relatively short element - just draw a line
                DrawLine(element,'#cccccc',alpha=0.1)



from setuptools import setup, find_packages

try:
    import pypandoc
    long_description = pypandoc.convert_file("README.md", "rst")
except ImportError:
    print ("Warning: pypandoc module not found, could not convert"
           " Markdown to reStructuredText." )
    long_description = ""


setup(
    name='pymadx',
    version='1.3',
    packages=find_packages(exclude=["docs", "tests", "obsolete"]),
    # Not sure how strict these need to be...
    install_requires=["matplotlib >= 1.7.1",
                      "numpy >= 1.3.0",
                      "pytransport"],
    # Some version of python2.7
    python_requires="==2.7.*",

    author='JAI@RHUL',
    author_email='laurie.nevay@rhul.ac.uk',
    description="Write MADX models and load MADX output.",
    long_description=long_description,
    url='https://bitbucket.org/jairhul/pymadx/',
    license='GPL3',
    keywords='madx accelerator twiss ptc',
)



"""
Analysis utilities for PTC data.
"""

import Ptc as _Ptc 
from  pymadx.Data import Tfs as _Tfs
import numpy as _np
import csv
import time

class PtcAnalysis(object):
    """
    Deprecated.

    Optical function calculation for PTC tracking data.

    This has be reimplemented and replaced by C++ implementation
    in rebdsim.

    """
    def __init__(self,ptcInput = None, ptcOutput = None) : 

        # Load input rays 
        if type(ptcInput) == str: 
            self.ptcInput = _Ptc.LoadInrays(ptcInput)
        else : 
            self.ptcInput = ptcInput 

        # Load output rays 
        if type(ptcOutput) == str:
            self.ptcOutput = _Tfs(ptcOutput)
        else : 
            self.ptcOutput = ptcOutput
    
    def SamplerLoop(self):
        #for segment in ptcOutput.segments:
        for isampler in range(len(self.ptcOutput.segments)):
            samplerData = self.ptcOutput.GetSegment(isampler+1)
            
            xrms  = samplerData.GetColumn('X').std()
            yrms  = samplerData.GetColumn('Y').std()
            pxrms = samplerData.GetColumn('PX').std()
            pyrms = samplerData.GetColumn('PY').std()

            print isampler+1, xrms, pxrms, yrms, pyrms

    def CalculateOpticalFunctions(self, output):
        """
        Calulates optical functions from a PTC output file
    
        output - the name of the output file
        """
        sampler  = []
        S        = []
        betx     = []
        bety     = []
        alphx    = []
        alphy    = []
        dispx    = []
        dispy    = []
        dispxp   = []
        dispyp   = []
        emittx   = []
        emitty   = []
        sigmax   = []
        sigmay   = []
        sigmaxp  = []
        sigmayp  = []
        sigmaxxp = []
        sigmayyp = []
        meanx    = []
        meany    = []
        meanxp   = []
        meanyp   = []
        Wgt      = []
        sigm_emx = []
        sigm_emy = []
        sigm_btx = []
        sigm_bty = []
        sigm_alx = []
        sigm_aly = []
        

        
        #for segment in ptcOutput.segments:
        nsegments = len(self.ptcOutput.segments)
        for isampler in range(1,len(self.ptcOutput.segments)+1):
            samplerData = self.ptcOutput.GetSegment(isampler)
            print 'segment:', isampler ,'/', nsegments
        
            x  = samplerData.GetColumn('X')
            y  = samplerData.GetColumn('Y') 
            xp = samplerData.GetColumn('PX')
            yp = samplerData.GetColumn('PY')
            s  = samplerData.GetColumn('S')
            E0 = samplerData.GetColumn('E')  #this is the specified beam  energy
            PT = samplerData.GetColumn('PT') # this is defined as pt= deltaE/(p0*c)
            E=E0*(1+PT) #This is the energy with a spread

            #Calculate sums
            s_s    = _np.sum(s)
            x_s    = _np.sum(x)
            y_s    = _np.sum(y)
            xp_s   = _np.sum(xp)
            yp_s   = _np.sum(yp)
            E_s    = _np.sum(E)
            EE_s   = _np.sum(E*E)
            xx_s   = _np.sum(x*x)
            xxp_s  = _np.sum(x*xp)
            xpxp_s = _np.sum(xp*xp)
            xpE_s  = _np.sum(xp*E)
            xE_s   = _np.sum(x*E)
            yy_s   = _np.sum(y*y)
            yyp_s  = _np.sum(y*yp)
            ypyp_s = _np.sum(yp*yp)
            ypE_s  = _np.sum(yp*E)
            yE_s   = _np.sum(y*E)

            wgt   = len(x) #this is a valid for ptc output as weight is always 1

            #normalise
            s_s    /= wgt
            x_s    /= wgt
            y_s    /= wgt
            xp_s   /= wgt
            yp_s   /= wgt
            E_s    /= wgt
            EE_s   /= wgt
            xx_s   /= wgt
            xxp_s  /= wgt
            xpxp_s /= wgt
            xpE_s  /= wgt
            xE_s   /= wgt
            yy_s   /= wgt
            yyp_s  /= wgt
            ypyp_s /= wgt
            ypE_s  /= wgt
            yE_s   /= wgt
            
            x_sv    = _np.sum(x-_np.mean(x))
            xx_sv   = _np.sum((x-_np.mean(x))*(x-_np.mean(x)))
            y_sv    = _np.sum(y-_np.mean(y))
            yy_sv   = _np.sum((y-_np.mean(y))*(y-_np.mean(y)))

            xp_sv   = _np.sum(xp-_np.mean(xp))
            xpxp_sv = _np.sum((xp-_np.mean(xp))*(xp-_np.mean(xp)))
            yp_sv   = _np.sum(yp-_np.mean(yp))
            ypyp_sv = _np.sum((yp-_np.mean(yp))*(yp-_np.mean(yp)))
            
            #Calculate variances/sigmas
            variance_x = (xx_sv - (x_sv * x_sv)/wgt)/wgt
            variance_y = (yy_sv - (y_sv * y_sv)/wgt)/wgt
            sigma_x    = _np.sqrt(variance_x)
            sigma_y    = _np.sqrt(variance_y)
            variance_xp = (xpxp_sv - (xp_sv * xp_sv)/wgt)/wgt
            variance_yp = (ypyp_sv - (yp_sv * yp_sv)/wgt)/wgt
            sigma_xp   = _np.sqrt(variance_xp)
            sigma_yp   = _np.sqrt(variance_yp)
        
            #Calculate means
            x_m  = x_s/wgt
            y_m  = y_s/wgt
            xp_m = xp_s/wgt
            yp_m = yp_s/wgt
        
            #Calculate sigmas
            sigma_x_xp = xxp_s - xx_s * xp_s;
            sigma_y_yp = yyp_s - y_s * yp_s;
            
            #####################
            #Calculate moments for error calculations
            
            mx_1_1  = _np.sum((x-x_m)*(xp-xp_m))/wgt
            mx_0_2  = _np.sum((xp-xp_m)**2)/wgt
            mx_2_0  = _np.sum((x-x_m)**2)/wgt
            mx_2_2  = _np.sum((x-x_m)**2*(xp-xp_m)**2)/wgt
            mx_1_3  = _np.sum((x-x_m)*(xp-xp_m)**3)/wgt
            mx_3_1  = _np.sum((x-x_m)**3*(xp-xp_m))/wgt
            mx_4_0  = _np.sum((x-x_m)**4)/wgt
            mx_0_4  = _np.sum((xp-xp_m)**4)/wgt

            my_1_1  = _np.sum((y-y_m)*(yp-yp_m))/wgt
            my_0_2  = _np.sum((yp-yp_m)**2)/wgt
            my_2_0  = _np.sum((y-y_m)**2)/wgt
            my_2_2  = _np.sum((y-y_m)**2*(yp-yp_m)**2)/wgt
            my_1_3  = _np.sum((y-y_m)*(yp-yp_m)**3)/wgt
            my_3_1  = _np.sum((y-y_m)**3*(yp-yp_m))/wgt
            my_4_0  = _np.sum((y-y_m)**4)/wgt
            my_0_4  = _np.sum((yp-yp_m)**4)/wgt
            
            
            #caculate higher moment covariances
            cov_vv_x       = -((-3+wgt)*mx_2_0**2)/((-1+wgt)*wgt)+mx_4_0/wgt
            cov_vv_xp      = -((-3+wgt)*mx_0_2**2)/((-1+wgt)*wgt)+mx_0_4/wgt
            cov_cc_xxp_xxp = -((-2+wgt)*mx_1_1**2)/((-1+wgt)*wgt)+(mx_0_2*mx_2_0)/((-1+wgt)*wgt)+mx_2_2/wgt
            cov_vc_x_xxp   = -((-3+wgt)*mx_1_1*mx_2_0)/((-1+wgt)*wgt)+mx_3_1/wgt
            cov_vc_xp_xxp  = -((-3+wgt)*mx_1_1*mx_0_2)/((-1+wgt)*wgt)+mx_1_3/wgt
            cov_vv_x_xp    = (2*mx_1_1**2)/((-1+wgt)*wgt)-(mx_0_2*mx_2_0)/wgt+mx_2_2/wgt

            cov_vv_y       = -((-3+wgt)*my_2_0**2)/((-1+wgt)*wgt)+my_4_0/wgt
            cov_vv_yp      = -((-3+wgt)*my_0_2**2)/((-1+wgt)*wgt)+my_0_4/wgt
            cov_cc_yyp_yyp = -((-2+wgt)*my_1_1**2)/((-1+wgt)*wgt)+(my_0_2*my_2_0)/((-1+wgt)*wgt)+my_2_2/wgt
            cov_vc_y_yyp   = -((-3+wgt)*my_1_1*my_2_0)/((-1+wgt)*wgt)+my_3_1/wgt
            cov_vc_yp_yyp  = -((-3+wgt)*my_1_1*my_0_2)/((-1+wgt)*wgt)+my_1_3/wgt
            cov_vv_y_yp    = (2*my_1_1**2)/((-1+wgt)*wgt)-(my_0_2*my_2_0)/wgt+my_2_2/wgt
 
            #Calculate the moments using the sums
            xx_s   -= x_s*x_s
            xxp_s  -= x_s*xp_s
            xpxp_s -= xp_s*xp_s
            yy_s   -= y_s*y_s  
            yyp_s  -= y_s*yp_s
            ypyp_s -= yp_s*yp_s

            #Calculate emittance
            emitt_x = _np.sqrt(xx_s*xpxp_s - xxp_s*xxp_s)
            emitt_y = _np.sqrt(yy_s*ypyp_s - yyp_s*yyp_s)

            #Calculate optical functions
            beta_x  =  xx_s  / emitt_x;
            beta_y  =  yy_s  / emitt_y;
            alph_x  = -xxp_s / emitt_x;
            alph_y  = -yyp_s / emitt_y;
            
            #if there is no energy spread it is expected that the dispersion would evaluate to inf, hence ignore
            #'divide by 0' runtime warning
            with _np.errstate(divide='ignore'):
                disp_x  = (xE_s  - (x_s  * E_s)) / (EE_s - (E_s * E_s))
                disp_xp = (xpE_s - (xp_s * E_s)) / (EE_s - (E_s * E_s))
                disp_y  = (yE_s  - (y_s  * E_s)) / (EE_s - (E_s * E_s))
                disp_yp = (ypE_s - (yp_s * E_s)) / (EE_s - (E_s * E_s))
            

            #error calculation
            d_emx_d_xx   = xpxp_s/(2*emitt_x)
            d_emx_d_xxp  = -xxp_s/emitt_x
            d_emx_d_xpxp = xx_s/(2*emitt_x)
            
            var_emitt_x = (d_emx_d_xx)**2*cov_vv_x+(d_emx_d_xxp)**2*cov_cc_xxp_xxp+(d_emx_d_xpxp)**2*cov_vv_xp
            var_emitt_x += 2*(d_emx_d_xx*d_emx_d_xxp)*cov_vc_x_xxp+2*(d_emx_d_xx*d_emx_d_xpxp)*cov_vv_x_xp
            var_emitt_x += 2*(d_emx_d_xxp*d_emx_d_xpxp)*cov_vc_xp_xxp
            
            sigma_emitt_x = _np.sqrt(var_emitt_x)

            d_emy_d_yy   = ypyp_s/(2*emitt_y)
            d_emy_d_yyp  = -yyp_s/emitt_y
            d_emy_d_ypyp = yy_s/(2*emitt_y)

            var_emitt_y = (d_emy_d_yy)**2*cov_vv_y+(d_emy_d_yyp)**2*cov_cc_yyp_yyp+(d_emy_d_ypyp)**2*cov_vv_yp
            var_emitt_y += 2*(d_emy_d_yy*d_emy_d_yyp)*cov_vc_y_yyp+2*(d_emy_d_yy*d_emy_d_ypyp)*cov_vv_y_yp
            var_emitt_y += 2*(d_emy_d_yyp*d_emy_d_ypyp)*cov_vc_yp_yyp
            
            sigma_emitt_y = _np.sqrt(var_emitt_y)
         
            d_btx_d_xx   = (xx_s*xpxp_s-2*xxp_s**2)/(2*(xx_s*xpxp_s - xxp_s*xxp_s)**(3./2.))
            d_btx_d_xxp  = (xx_s*xxp_s)/((xx_s*xpxp_s - xxp_s*xxp_s)**(3./2.))
            d_btx_d_xpxp = -(xx_s**2)/(2*(xx_s*xpxp_s - xxp_s*xxp_s)**(3./2.))

            var_beta_x = (d_btx_d_xx)**2*cov_vv_x+(d_btx_d_xxp)**2*cov_cc_xxp_xxp+(d_btx_d_xpxp)**2*cov_vv_xp
            var_beta_x+= 2*(d_btx_d_xx*d_btx_d_xxp)*cov_vc_x_xxp+2*(d_btx_d_xx*d_btx_d_xpxp)*cov_vv_x_xp
            var_beta_x+= 2*(d_btx_d_xxp*d_btx_d_xpxp)*cov_vc_xp_xxp
            
            sigma_beta_x = _np.sqrt(var_beta_x)

            d_bty_d_yy   = (yy_s*ypyp_s-2*yyp_s**2)/(2*(yy_s*ypyp_s - yyp_s*yyp_s)**(3./2.))
            d_bty_d_yyp  = (yy_s*yyp_s)/((yy_s*ypyp_s - yyp_s*yyp_s)**(3./2.))
            d_bty_d_ypyp = -(yy_s**2)/(2*(yy_s*ypyp_s - yyp_s*yyp_s)**(3./2.))

            var_beta_y = (d_bty_d_yy)**2*cov_vv_y+(d_bty_d_yyp)**2*cov_cc_yyp_yyp+(d_bty_d_ypyp)**2*cov_vv_yp
            var_beta_y+= 2*(d_bty_d_yy*d_bty_d_yyp)*cov_vc_y_yyp+2*(d_bty_d_yy*d_bty_d_ypyp)*cov_vv_y_yp
            var_beta_y+= 2*(d_bty_d_yyp*d_bty_d_ypyp)*cov_vc_yp_yyp
            
            sigma_beta_y = _np.sqrt(var_beta_y)

            d_alx_d_xx   = -(xxp_s*xpxp_s)/(2*(xx_s*xpxp_s - xxp_s*xxp_s)**(3./2.))
            d_alx_d_xxp  = (xx_s*xpxp_s)/((xx_s*xpxp_s - xxp_s*xxp_s)**(3./2.))
            d_alx_d_xpxp = -(xx_s*xxp_s)/(2*(xx_s*xpxp_s - xxp_s*xxp_s)**(3./2.))

            var_alph_x = (d_alx_d_xx)**2*cov_vv_x+(d_alx_d_xxp)**2*cov_cc_xxp_xxp+(d_alx_d_xpxp)**2*cov_vv_xp
            var_alph_x+= 2*(d_alx_d_xx*d_alx_d_xxp)*cov_vc_x_xxp+2*(d_alx_d_xx*d_alx_d_xpxp)*cov_vv_x_xp
            var_alph_x+= 2*(d_alx_d_xxp*d_alx_d_xpxp)*cov_vc_xp_xxp
            
            sigma_alph_x = _np.sqrt(var_alph_x)

            
            d_aly_d_yy   = -(yyp_s*ypyp_s)/(2*(yy_s*ypyp_s - yyp_s*yyp_s)**(3./2.))
            d_aly_d_yyp  = (yy_s*ypyp_s)/((yy_s*ypyp_s - yyp_s*yyp_s)**(3./2.))
            d_aly_d_ypyp = -(yy_s*yyp_s)/(2*(yy_s*ypyp_s - yyp_s*yyp_s)**(3./2.))

            var_alph_y = (d_aly_d_yy)**2*cov_vv_y+(d_aly_d_yyp)**2*cov_cc_yyp_yyp+(d_aly_d_ypyp)**2*cov_vv_yp
            var_alph_y+= 2*(d_aly_d_yy*d_aly_d_yyp)*cov_vc_y_yyp+2*(d_aly_d_yy*d_aly_d_ypyp)*cov_vv_y_yp
            var_alph_y+= 2*(d_aly_d_yyp*d_aly_d_ypyp)*cov_vc_yp_yyp
            
            sigma_alph_y = _np.sqrt(var_alph_y)
            
            """
            print "mx_2_0= ",mx_2_0,", mx_0_2= ",mx_0_2,", mx_1_1= ",mx_1_1
            print "mx_4_0= ",mx_4_0,", mx_0_4= ",mx_0_4
            print "mx_3_1= ",mx_3_1,", mx_1_3= ",mx_1_3,", mx_2_2= ",mx_2_2
            print "my_2_0= ",my_2_0,", my_0_2= ",my_0_2,", my_1_1= ",my_1_1
            print "my_4_0= ",my_4_0,", my_0_4= ",my_0_4
            print "my_3_1= ",my_3_1,", my_1_3= ",my_1_3,", my_2_2= ",my_2_2

            print "cov_vv_x= ", cov_vv_x,", cov_vv_xp= ",cov_vv_xp,", cov_cc_xxp_xxp= ",cov_cc_xxp_xxp
            print "cov_vc_x_xxp= ",cov_vc_x_xxp,", cov_vc_xp_xxp= ",cov_vc_xp_xxp,", cov_vv_x_xp=",cov_vv_x_xp
            print "cov_vv_y= ",cov_vv_y,", cov_vv_yp= ",cov_vv_yp,", cov_cc_yyp_yyp= ",cov_cc_yyp_yyp
            print "cov_vc_y_yyp= ",cov_vc_y_yyp,", cov_vc_yp_yyp= ",cov_vc_yp_yyp,", cov_vv_y_yp=",cov_vv_y_yp
           
            print "dedx= ",d_emx_d_xx," ,dedxp= ",d_emx_d_xpxp,", dedxxp= ",d_emx_d_xxp
            print "dbdx= ",d_btx_d_xx,",dbdxp= ",d_btx_d_xpxp,", dbdxxp= ",d_btx_d_xxp
            print "dadx= ",d_alx_d_xx,",dadxp= ",d_alx_d_xpxp,", dadxxp= ",d_alx_d_xxp
            print "dedy= ",d_emy_d_yy," ,dedyp= ",d_emy_d_ypyp,", dedyyp= ",d_emy_d_yyp
            print "dbdy= ",d_bty_d_yy,",dbdyp= ",d_bty_d_ypyp,", dbdyyp= ",d_bty_d_yyp
            print "dady= ",d_aly_d_yy,",dadyp= ",d_aly_d_ypyp,", dadyyp= ",d_aly_d_yyp

            print "var_emitt_x= ",var_emitt_x,", var_beta_x= ",var_beta_x,", var_alph_x= ",var_alph_x
            print "var_emitt_y= ",var_emitt_y,", var_beta_y= ",var_beta_y,", var_alph_y= ",var_alph_y

            print "Emiitance_x= ",emitt_x," Sigma_Emittance_x = ",sigma_emitt_x, " Frac error= ", sigma_emitt_x/emitt_x 
            print "Beta_x= ",beta_x," Sigma_Beta_x = ",sigma_beta_x, " Frac error= ", sigma_beta_x/beta_x
            print "Alpha_x= ",alph_x," Sigma_Alpha_x = ",sigma_alph_x, " Frac error= ", sigma_alph_x/alph_x
            """

            sampler.append(isampler)
            S.append(s_s)            
            betx.append(beta_x)
            bety.append(beta_y)
            alphx.append(alph_x)
            alphy.append(alph_y)
            dispx.append(disp_x)
            dispy.append(disp_y)
            dispxp.append(disp_xp)
            dispyp.append(disp_yp)
            emittx.append(emitt_x)
            emitty.append(emitt_y)
            sigmax.append(sigma_x)
            sigmay.append(sigma_y)
            sigmaxp.append(sigma_xp)
            sigmayp.append(sigma_yp)
            sigmaxxp.append(sigma_x_xp)
            sigmayyp.append(sigma_y_yp)
            meanx.append(x_m)
            meany.append(y_m)
            meanxp.append(xp_m)
            meanyp.append(yp_m)
            Wgt.append(wgt)
            sigm_emx.append(sigma_emitt_x)
            sigm_emy.append(sigma_emitt_y)
            sigm_btx.append(sigma_beta_x)
            sigm_bty.append(sigma_beta_y)
            sigm_alx.append(sigma_alph_x)
            sigm_aly.append(sigma_alph_y)

        #prepare header    
        header = ['Segment','S[m]','Beta_x[m]','Beta_y[m]','Alph_x','Alph_y']
        header.extend(['Disp_x','Disp_xp','Disp_y','Disp_yp'])
        header.extend(['Emitt_x','Emitt_y'])
        header.extend(['Sigma_x[m]','Sigma_y[m]','Sigma_xp[rad]','Sigma_yp[rad]'])
        header.extend(['Sigma_x_xp[m*rad]','Sigma_y_yp[m*rad]'])
        header.extend(['Mean_x[m]','Mean_y[m]','Mean_xp[rad]','Mean_yp[rad]','Wgt'])
        header.extend(['Sigma_emitt_x','Sigma_emitt_y','Sigma_beta_x','Sigma_beta_y'])
        header.extend(['Sigma_alph_x','Sigma_alph_y'])


        #prepare optical function arrays for writing to file

        for i in range(len(S)):
            """
            S[i]=float("{0:.4e}".format(S[i]))
            betx[i]=float("{0:.4e}".format(betx[i]))
            bety[i]=float("{0:.4e}".format(bety[i]))
            alphx[i]=float("{0:.4e}".format(alphx[i]))
            alphy[i]=float("{0:.4e}".format(alphy[i]))
            dispx[i]=float("{0:.4e}".format(dispx[i]))
            dispy[i]=float("{0:.4e}".format(dispy[i]))
            dispxp[i]=float("{0:.4e}".format(dispxp[i]))
            dispyp[i]=float("{0:.4e}".format(dispyp[i]))
            emittx[i]=float("{0:.4e}".format(emittx[i]))
            emitty[i]=float("{0:.4e}".format(emitty[i]))
            sigmax[i]=float("{0:.4e}".format(sigmax[i]))
            sigmay[i]=float("{0:.4e}".format(sigmay[i]))
            sigmaxp[i]=float("{0:.4e}".format(sigmaxp[i]))
            sigmayp[i]=float("{0:.4e}".format(sigmayp[i]))
            sigmaxxp[i]=float("{0:.4e}".format(sigmaxxp[i]))
            sigmayyp[i]=float("{0:.4e}".format(sigmayyp[i]))
            meanx[i]=float("{0:.4e}".format(meanx[i]))
            meany[i]=float("{0:.4e}".format(meany[i]))
            meanxp[i]=float("{0:.4e}".format(meanxp[i]))
            meanyp[i]=float("{0:.4e}".format(meanyp[i]))
            Wgt[i]=float("{0:.4e}".format(Wgt[i]))
            """

        with open(output,'w') as ofile:        
            writer=csv.writer(ofile, delimiter='\t',lineterminator='\n',)
            timestamp = time.strftime("%Y/%m/%d-%H:%M:%S")
            writer.writerow(['# ','Optical functions from PTC output', timestamp])
            writer.writerow(header)
            for i in range(len(S)):
                row = [sampler[i],S[i],betx[i],bety[i],alphx[i],alphy[i],dispx[i],dispy[i]]
                row.extend([dispxp[i],dispyp[i],emittx[i],emitty[i]])
                row.extend([sigmax[i],sigmay[i],sigmaxp[i],sigmayp[i],sigmaxxp[i],sigmayyp[i]])
                row.extend([meanx[i],meany[i],meanxp[i],meanyp[i],Wgt[i]])
                row.extend([sigm_emx[i],sigm_emy[i],sigm_btx[i],sigm_bty[i]])
                row.extend([sigm_alx[i],sigm_aly[i]])
                writer.writerow(row)

        self.opticalFunctions = {
            'Segment'       : sampler,
            'S'             : S,
            'beta_x'        : betx,
            'beta_y'        : bety,
            'alph_x'        : alphx,
            'alph_y'        : alphy,
            'disp_x'        : dispx,
            'disp_xp'       : dispxp,
            'disp_y'        : dispy,
            'disp_yp'       : dispyp,
            'emitt_x'       : emittx,
            'emitt_y'       : emitty,
            'sigma_x'       : sigmax,
            'sigma_y'       : sigmay,
            'sigma_xp'      : sigmaxp,
            'sigma_yp'      : sigmayp,
            'sigma_x_xp'    : sigmaxxp,
            'sigma_y_yp'    : sigmayyp,
            'mean_x'        : meanx,
            'mean_y'        : meany,
            'mean_xp'       : meanxp,
            'mean_yp'       : meanyp,
            'wgt'           : Wgt,
            'sigma_emitt_x' : sigm_emx,
            'sigma_emitt_y' : sigm_emy,
            'sigma_beta_x'  : sigm_btx,
            'sigma_beta_y'  : sigm_bty,
            'sigma_alph_x'  : sigm_alx,
            'sigma_alph_y'  : sigm_aly
            }
        



"""
Functions for comparing the optical distributions of two
BDSIM models.

Functions for plotting the individual optical functions, and an
eighth, helper function ''compare_all_optics``, to plot display all
seven in one go.
"""
import datetime as _datetime
import matplotlib.pyplot as _plt
from matplotlib.backends.backend_pdf import PdfPages as _PdfPages
import os.path as _path

import pymadx.Data as _Data
import pymadx.Plot as _Plot

# Predefined lists of tuples for making the standard plots,
# format = (optical_var_name, optical_var_error_name, legend_name)

_BETA = [("BETX", r'$\beta_{x}$'),
         ("BETY", r'$\beta_{y}$')]

_ALPHA = [("ALFX", r"$\alpha_{x}$"),
          ("ALFY", r"$\alpha_{y}$")]

_DISP = [("DXBETA", r"$D_{x}$ / $\beta$"),
         ("DYBETA", r"$D_{y}$ / $\beta$")]

_DISP_P = [("DPXBETA", r"$D_{p_{x}}$"),
           ("DPYBETA", r"$D_{p_{y}}$")]

_SIGMA = [("SIGMAX", r"$\sigma_{x}$"),
          ("SIGMAY", r"$\sigma_{y}$")]

_SIGMA_P = [("SIGMAXP", r"$\sigma_{xp}$"),
            ("SIGMAYP", r"$\sigma_{yp}$")]

_MEAN = [("X", r"$\bar{x}$"),
         ("Y", r"$\bar{y}$")]


def _parse_tfs_input(tfs_in, name):
    """Return tfs_in as a Tfs instance, which should either be a path
    to a TFS file or a Tfs instance, and in either case, generate a
    name if None is provided, and return that as well."""
    if isinstance(tfs_in, basestring):
        if not _path.isfile(tfs_in):
            raise IOError("file \"{}\" not found!".format(tfs_in))
        name = (_path.splitext(_path.basename(tfs_in))[0]
                if name is None else name)
        return _Data.Tfs(tfs_in), name
    try:
        name = tfs_in.filename if name is None else name
        return tfs_in, name
    except AttributeError:
        raise TypeError(
            "Expected Tfs input is neither a "
            "file path nor a Tfs instance: {}".format(tfs_in))


# use closure to avoid tonnes of boilerplate code as happened with
# MadxBdsimComparison.py
def _make_plotter(plot_info_tuples, x_label, y_label, title):
    def f_out(first, second, first_name=None, second_name=None, **kwargs):
        """first and second should be tfs files."""
        first, first_name = _parse_tfs_input(first, first_name)
        second, second_name = _parse_tfs_input(second, second_name)

        plot = _plt.figure(title, figsize=(9,5), **kwargs)
        # Loop over the variables in plot_info_tuples and draw the plots.
        for var, legend_name in plot_info_tuples:
            _plt.plot(first.GetColumn('S'), first.GetColumn(var),
                      label="{}: {}".format(legend_name, first_name), **kwargs)
            _plt.plot(second.GetColumn('S'), second.GetColumn(var),
                      label="{}: {}".format(legend_name, second_name), **kwargs)

        # Set axis labels and draw legend
        axes = _plt.gcf().gca()
        axes.set_ylabel(y_label)
        axes.set_xlabel(x_label)
        axes.legend(loc='best')

        _Plot.AddMachineLatticeToFigure(plot, first)
        _plt.show(block=False)
        return plot
    return f_out

PlotBeta   = _make_plotter(_BETA,    "S / m", r"$\beta_{x,y}$ / m",      "Beta")
PlotAlpha  = _make_plotter(_ALPHA,   "S / m", r"$\alpha_{x,y}$ / m",     "Alpha")
PlotDisp   = _make_plotter(_DISP,    "S / m", r"$D_{x,y} / m$",          "Dispersion")
PlotDispP  = _make_plotter(_DISP_P,  "S / m", r"$D_{p_{x},p_{y}}$ / m",  "Momentum_Dispersion")
PlotSigma  = _make_plotter(_SIGMA,   "S / m", r"$\sigma_{x,y}$ / m",     "Sigma")
PlotSigmaP = _make_plotter(_SIGMA_P, "S / m", r"$\sigma_{xp,yp}$ / rad", "SigmaP")
PlotMean   = _make_plotter(_MEAN,    "S / m", r"$\bar{x}, \bar{y}$ / m", "Mean")


def MADXVsMADX(first, second, first_name=None,
               second_name=None, saveAll=True, 
               outputFileName=None, **kwargs):
    """
    Display all the optical function plots for the two input optics files.
    """
    figures = [
    PlotBeta(first, second, first_name=first_name,
             second_name=second_name, **kwargs),
    PlotAlpha(first, second, first_name=first_name,
              second_name=second_name, **kwargs),
    PlotDisp(first, second, first_name=first_name,
             second_name=second_name, **kwargs),
    PlotDispP(first, second, first_name=first_name,
              second_name=second_name, **kwargs),
    PlotSigma(first, second, first_name=first_name,
              second_name=second_name, **kwargs),
    PlotSigmaP(first, second, first_name=first_name,
               second_name=second_name, **kwargs),
    PlotMean(first, second, first_name=first_name,
             second_name=second_name, **kwargs),
    ]

    if saveAll:
        if outputFileName is not None:
            output_filename = outputFileName
        else:
            output_filename = "optics-report.pdf"

        with _PdfPages(output_filename) as pdf:
            for figure in figures:
                pdf.savefig(figure)
            d = pdf.infodict()
            d['Title'] = "{} VS {} Optical Comparison".format(first_name, second_name)
            d['CreationDate'] = _datetime.datetime.today()
        print "Written ", output_filename



import pymadx.Beam as _Beam
import pymadx.Builder as _Builder
import numpy as _np
import re as _re
import csv

import pymadx.Data as _Data

def TfsToPtc(inputfile,outputfilename, ptcfile, startname=None,
             stopname=None,ignorezerolengthitems=True,samplers='all',
             beampiperadius=0.2,beam=True,ptctrackaperture=[]):
    """
    Prepare a madx model for PTC from a Tfs file containing the full
    twiss output from MADX.

    """

    madx = _Data.CheckItsTfs(inputfile)
    ptcfilename = ptcfile
        
    nitems = madx.nitems
    opencollimatorsetting = beampiperadius

    # data structures for checks
    angtot = 0.0
    lentot = 0.0
    lldiff = []
    dldiff = {}
    itemsomitted = []

    a = _Builder.Machine()
    
    #prepare index for iteration:
    if startname == None:
        startindex = 0
    elif type(startname) == int:
        startindex = startname
    else:
        startindex = madx.IndexFromName(startname)
    if stopname   == None:
        stopindex = nitems #this is 1 larger, but ok as range will stop at n-step -> step=1, range function issue
    elif type(stopname) == int:
        stopindex = stopname
    else:
        stopindex  = madx.IndexFromName(stopname)
    if stopindex <= startindex:
        print 'stopindex <= startindex'
        stopindex = startindex + 1

    try:
        lindex          = madx.ColumnIndex('L')
        angleindex      = madx.ColumnIndex('ANGLE')
        ksIindex        = madx.ColumnIndex('KSI')
        k0lindex        = madx.ColumnIndex('K0L')
        k0slindex       = madx.ColumnIndex('K0SL')
        k1lindex        = madx.ColumnIndex('K1L')
        k2lindex        = madx.ColumnIndex('K2L')
        k3lindex        = madx.ColumnIndex('K3L')
        k4lindex        = madx.ColumnIndex('K4L')
        k5lindex        = madx.ColumnIndex('K5L')
        k6lindex        = madx.ColumnIndex('K6L')
        k1slindex       = madx.ColumnIndex('K1SL')
        k2slindex       = madx.ColumnIndex('K2SL')
        k3slindex       = madx.ColumnIndex('K3SL')
        k4slindex       = madx.ColumnIndex('K4SL')
        k5slindex       = madx.ColumnIndex('K5SL')
        k6slindex       = madx.ColumnIndex('K6SL')
        tiltindex       = madx.ColumnIndex('TILT')
        tindex          = madx.ColumnIndex('KEYWORD')
        alphaxindex     = madx.ColumnIndex('ALFX')
        alphayindex     = madx.ColumnIndex('ALFY')
        betaxindex      = madx.ColumnIndex('BETX')
        betayindex      = madx.ColumnIndex('BETY')
        vkickangleindex = madx.ColumnIndex('VKICK')
        hkickangleindex = madx.ColumnIndex('HKICK')
        e1index         = madx.ColumnIndex('E1')
        e2index         = madx.ColumnIndex('E2')
        fintindex       = madx.ColumnIndex('FINT')
        fintxindex      = madx.ColumnIndex('FINTX')
        hgapindex       = madx.ColumnIndex('HGAP')

    except ValueError:
        print 'Missing columns from tfs file - insufficient information to convert file'
        print 'Required columns : L, ANGLE, KSI, K1L...K6L, K1SL...K6SL, TILT, KEYWORD, ALFX, ALFY, BETX, BETY, VKICK, HKICK'
        print 'Given columns    : '
        print madx.columns

        
    # iterate through input file and construct machine
    for i in range(startindex,stopindex):
        name = madx.sequence[i]
        #remove special characters like $, % etc 'reduced' name - rname
        rname = _re.sub('[^a-zA-Z0-9_]+','',name) #only allow alphanumeric characters and '_'
        t     = madx.data[name][tindex]
        l     = madx.data[name][lindex]
        ang   = madx.data[name][angleindex]
        if l <1e-9:
            zerolength = True
        else:
            zerolength = False

        if zerolength and ignorezerolengthitems:
            itemsomitted.append(name)
            continue #this skips the rest of the loop as we're ignoring this item

        kws = {} # element-specific keywords
        tilt = madx.data[name][tiltindex]
        if tilt != 0:
            kws['tilt'] = tilt
        
        e1    = madx.data[name][e1index]
        e2    = madx.data[name][e2index]
        fint  = madx.data[name][fintindex]
        fintx = madx.data[name][fintxindex]
        hgap  = madx.data[name][hgapindex]
        k1l   = madx.data[name][k1lindex]

        if k1l:
            kws['k1'] = k1l / l

        

        if t == 'DRIFT':
            a.AddDrift(rname,l,**kws)

        elif t == 'QUADRUPOLE':
            k1 = madx.data[name][k1lindex] / l
            a.AddQuadrupole(rname,l,**kws)

        elif t == 'SEXTUPOLE':
            k2 = madx.data[name][k2lindex] / l
            a.AddSextupole(rname,l,k2=k2,**kws)

        elif t == 'OCTUPOLE':
            k3 = madx.data[name][k3lindex] / l
            a.AddOctupole(rname,l,k3=k3,**kws)

        elif t == 'SBEND':
            kws['e1'] = e1
            kws['fint'] = fint
            kws['e2'] = e2

            if fintx == -1:
                if fint:
                    kws['fintx'] = fint
                else:
                    kws['fintx'] = 0
            else:
                kws['fintx'] = fintx

            kws['hgap'] = hgap

            angle = madx.data[name][angleindex]
            a.AddDipole(rname,category='sbend',length=l,angle=angle,**kws)

        elif t == 'RBEND':
            kws['e1'] = e1
            kws['fint'] = fint

            kws['e2'] = e2

            if fintx == -1:
                if fint:
                    kws['fintx'] = fint
                else:
                    kws['fintx'] = 0
            else:
                kws['fintx'] = fintx

            kws['hgap'] = hgap

            angle = madx.data[name][angleindex]
            a.AddDipole(rname,category='rbend',length=l,angle=angle,**kws)

        elif t == 'MARKER':
            angle = madx.data[name][angleindex]
            a.AddMarker(rname, **kws)

        elif t == 'MULTIPOLE':
            kn0l  = madx.data[name][k0lindex]
            kn1l  = madx.data[name][k1lindex]
            kn2l  = madx.data[name][k2lindex]
            kn3l  = madx.data[name][k3lindex]
            kn4l  = madx.data[name][k4lindex]
            kn5l  = madx.data[name][k5lindex]
            kn6l  = madx.data[name][k6lindex]
            kn0sl = madx.data[name][k0slindex]
            kn1sl = madx.data[name][k1slindex]
            kn2sl = madx.data[name][k2slindex]
            kn3sl = madx.data[name][k3slindex]
            kn4sl = madx.data[name][k4slindex]
            kn5sl = madx.data[name][k5slindex]
            kn6sl = madx.data[name][k6slindex]
            
            a.AddMultipole(rname, knl=(kn0l, kn1l, kn2l, kn3l, kn4l, kn5l, kn6l), ksl=(kn0sl, kn1sl, kn2sl, kn3sl, kn4sl, kn5sl, kn6sl),**kws)

        else:
            print 'MadxTfs2Ptc> unknown element type: ',t,' for element named: ',name
            if not zerolength:
                print('MadxTfs2Ptc> replacing with drift')
                a.AddDrift(rname,l)
            else:
                pass

    a.AddSampler(samplers)
    a.AddPTCTrackAperture(ptctrackaperture)
    # Make beam file 
    if beam: 
        b = MadxTfsToPtcBeam(madx, ptcfilename, startname)
        a.AddBeam(b)

    a.Write(outputfilename)

    return a

def MadxTfsToPtcBeam(tfs, ptcfilename,  startname=None):
    if startname == None:
        startindex = 0
    elif type(startname) == int:
        startindex = startname
    else:
        startindex = tfs.IndexFromName(startname)

    #MADX defines parameters at the end of elements so need to go 1 element
    #back if we can.

    if startindex > 0:
        startindex -= 1
    
    energy   = float(tfs.header['ENERGY'])
    gamma    = float(tfs.header['GAMMA'])
    particle = str.lower(tfs.header['PARTICLE'])
    ex       = tfs.header['EX']
    ey       = tfs.header['EY']
    sigmae   = float(tfs.header['SIGE'])
    sigmat   = float(tfs.header['SIGT'])

    data     = tfs.GetRowDict(tfs.sequence[startindex])

    beam  = _Beam.Beam(particle, energy,'ptc', emitx=ex, emity=ey, sigmaE=sigmae)

    beam.SetDistribFileName(ptcfilename) 

    return beam
